/**
 ** pif_Iterator
 **   Version: 1.02.20191216
 **   Release Date: December 16, 2019
 **
 ***************************************************************************************************
 ***************************************************************************************************

pif_Iterator is a library that implements an Iterator (/pif_Iterator) protocol. Iterators are an aid
used to iterate incrementally through data of various sorts, and are useful especially in traversing
data structures.

 +----------------------------------------------------+
 |                                                    |
 |  Contents                                          |
 |   1. License                                       |
 |   2. Introduction                                  |
 |   3. The Iterator Protocol                         |
 |   4. Preprocessors                                 |
 |   5. The /pif_Iterator protocol class              |
 |   6. The /pif_Iterator/ForwardListIterator class   |
 |   7. The /pif_Iterator/BackwardListIterator class  |
 |   8. The /pif_Iterator/RangeIterator class         |
 |   9. The /pif_Iterator/MethodIterator class        |
 |  10. Release Notes                                 |
 |                                                    |
 +----------------------------------------------------+

   1. License
  ----------------------------------------------------

  The code for pif_Iterator is released under the MIT License. Refer to "+License.dm" for the full
  text of the license.

  2. Introduction
  ----------------------------------------------------

  pif_Iterator implements a 'protocol class' for Iterator objects. By 'protocol class', I mean that
  the /pif_Iterator class defines seven methods (three of which are optional) to be used by
  subclasses to allow programmers to incrementally iterate through some set of data, whether it is
  from a data structure or streaming from another source or randomly generate as the sequence is
  iterated through.


  3. The Iterator Protocol
  ----------------------------------------------------

  The "pif_Iterator class.dm" file defines the pif_Iterator protocol class and details the protocol
  itself. Refer to this file for the protocol in full detail, but an abbreivation of it is here.

  A particular Iterator class should be designed to only accept and 'understand' data from a given
  source. That is, there should be no "swiss-army knife" Iterator designed to work with any and
  every possible source of data. Make a given Iterator class do one thing and do it well.

  The protocol consists of seven methods, along with the constructor. Their behavior is as follows.

    1. The Iterator class constructor should accept an argument indicating the data source to read
       from.

    2. The Iterator Initialize() method should be called to begin iterating through data from the
       source provided to the constructor.

    3. The Next() method moves to the 'next' piece of data from the data source, and returns either
       the new position or 0, if the end of the data has been reached. It may optionally take an
       argument indicating a position to move to.

    4. The Look() method gives the data from the source data at the current position.

    5. The Done() method returns true when the end of the data has been reached, and false
       otherwise.

    6. The Step() method moves the current position forward or backwards by a certain amount, as
       indicated by its argument. This method is optional.

    7. The Peek() method looks at the data in a position forward or backwards by a certain amount,
       as indicated by its argument. This method is optional.

    8. The Position() method indicates the current position in the data. If the data has no
       meaningful position, it may also be used as a counter, to indicate how much data has been
       moved through.

  4. Preprocessors
  ----------------------------------------------------

  By default, the namespace of this library is prefixed, i.e. /pif_Iterator has the prefix _pif on
  it. This can be removed by either,

    1. Defining PIF_NOPREFIX_ITERATOR, which will remove the prefix from only this library.
    2. Defining PIF_NOPREFIX_GENERAL, which will remove the prefix from this library and all others
       of mine that in the future will have similar prefixing.

  As expected, the unprefixed namespace is /Iterator.

  There is additionally a PIF_ITERATOR_INCLUDED flag defined, to allow for the optional inclusion of
  code depending on the presence of this library.

  5. The /pif_Iterator protocol class.
  ----------------------------------------------------

    +-------------------------------+
    |  Parent Class: /datum         |
    |  Namespace: /pif_Iterator     |
    |  File: pif_Iterator class.dm  |
    +-------------------------------+

  The /pif_Iterator class is used to implement the Iterator protocol. Refer to "pif_Iterator.dm" for
  complete details on the class and protocol.

  6. The /pif_Iterator/ForwardListIterator class
  ----------------------------------------------------

    +--------------------------------------+
    |  Parent Class: /pif_Iterator         |
    |  Namespace: /pif_Iterator            |
    |  File: ForwardListIterator class.dm  |
    +--------------------------------------+

  The /ForwardListIterator class is used to iterate through a /list object from beginning to end.
  That is, it iterates forward through the list.

    New()
    --------------------

    Arguments.

      A /list object. If the argument is not a /list object, or if none is provided, an exception is
      thrown.

    Behavior.

      Stores the /list object for iteration.

    Exceptions.

      Exception is thrown when no arguments are provided or the provided argument is not a /list
      object.

    Done()
    --------------------

    Returns.

      True if the Iterator object has reached the end of the data, and false otherwise.

    Initialize()
    --------------------

    Arguments.

      An optional numeric argument that indicates where the Iterator object begins iteration. It
      defaults to 1.

    Behavior.

      Sets up the Iterator to begin iterating through the list at the given position.

    Returns.

      1 on success, 0 on failure. Failure comes from an invalid argument.

    Look()
    --------------------

    Returns.

      The data at the current position, or null if the iterator is not initialized or has reached
      the end of the list.

    Next()
    --------------------

    Arguments.

      An optional numeric argument to indicate the position the Iterator should move to next.
      Defaults to null.

    Behavior.

      Will either move to the next position in the list, or the position specified by the argument
      if it is provided.

    Returns.

      The numeric value of the new position, or 0 if the end of the list has been reached.

    Peek()
    --------------------

    Arguments.

      An optional numeric argument to indicate the position to read from. Defaults to 1.

    Returns.

      The value at the specified position of the list, or null if it is beyond the bounds of the
      list.

    Position()
    --------------------

    Returns.

      The numeric value of the current position of the list.

    Step()
    --------------------

    Arguments.

      An optional numeric argument to indicate how many positions forward (if positive) or backwards
      (if negative) to move. By default, this amount is 1.

    Behavior.

      Moves the position of the list the specified amount.

    Returns.

      The numeric value of the new position, or 0 if the position moves out of bounds of the list.

    Note: Step(1) and Step() (i.e., Step with a default argument) have the same exact behavior as
          Next().

  7. The /pif_Iterator/BackwardListIterator class
  ----------------------------------------------------

    +---------------------------------------------------+
    |  Parent Class: /pif_Iterator/ForwardListIterator  |
    |  Namespace: /pif_Iterator                         |
    |  File: BackwardListIterator class.dm              |
    +---------------------------------------------------+

  The /BackwardListIterator class is used to iterate through a /list object from the last position
  to the first. That is, it iterates backwards through the list.

    New()
    --------------------

    Arguments.

      A /list object. If the argument is not a /list object, or if none is provided, an exception is
      thrown.

    Behavior.

      Stores the /list object for iteration.

    Exceptions.

      Exception is thrown when no arguments are provided or the provided argument is not a /list
      object.

    Done()
    --------------------

    Returns.

      True if the Iterator object has reached the end of the data, and false otherwise.

    Initialize()
    --------------------

    Arguments.

      An optional numeric argument that indicates where the Iterator object begins iteration. It
      defaults to 1.

    Behavior.

      Sets up the Iterator to begin iterating through the list at the given position.

    Returns.

      1 on success, 0 on failure. Failure comes from an invalid argument.

    Look()
    --------------------

    Returns.

      The data at the current position, or null if the iterator is not initialized or has reached
      the end of the list.

    Next()
    --------------------

    Arguments.

      An optional numeric argument to indicate the position the Iterator should move to next.
      Defaults to null.

    Behavior.

      Will either move to the next position in the list, or the position specified by the argument
      if it is provided.

    Returns.

      The numeric value of the new position, or 0 if the end of the list has been reached.

    Peek()
    --------------------

    Arguments.

      An optional numeric argument to indicate the position to read from. Defaults to 1.

    Returns.

      The value at the specified position of the list, or null if it is beyond the bounds of the
      list.

    Position()
    --------------------

    Returns.

      The numeric value of the current position of the list.

    Step()
    --------------------

    Arguments.

      An optional numeric argument to indicate how many positions forward (if positive) or backwards
      (if negative) to move. By default, this amount is 1.

    Behavior.

      Moves the position of the list the specified amount.

    Returns.

      The numeric value of the new position, or 0 if the position moves out of bounds of the list.

    Note: Step(1) and Step() (i.e., Step with a default argument) have the same exact behavior as
          Next().

  8. The /pif_Iterator/RangeIterator class
  ----------------------------------------------------

    +--------------------------------+
    |  Parent Class: /pif_Iterator   |
    |  Namespace: /pif_Iterator      |
    |  File: RangeIterator class.dm  |
    +--------------------------------+

  The RangeIterator class is used to iterate through a numeric range specified by the programmer,
  with an optional step size (defaulting to 1). While the Iterator will always output the starting
  point as a value, it will only output the closest it can get to end without going over it. Thus,
  whether the end point is included depends on whether there is a natural number N such that end =
  start + N*step.

    New()
    --------------------

    Arguments.

      Takes three arguments, one optional.

      1. The start position for the range.
      2. The end position for the range.
      3. An optional step size for the range, defaulting to 1. Must be non-zero, or an exception
         will be thrown.

    Behavior.

      Instantiates the RangeIterator object with the provided range data.

    Exceptions.

      An exception is thrown if the provided step size is zero.

    Done()
    --------------------

    Returns.

      True, if the iterator has reached the end of the range. False otherwise.

    Initialize()
    --------------------

    Behavior.

      Sets up the iterator to start iterating over the range.

    Returns.

      1.

    Look()
    --------------------

    Returns.

      The current location in the range.

    Next()
    --------------------

    Behavior.

      Moves to the next position in the data.

    Returns.

      The next position, or 0 if the end of the range has been reached.

    Position()
    --------------------

    Returns.

      The current position in the range. Essentially, a counter of how many times the Iterator
      has moved through the data.

  9. The /pif_Iterator/MethodIterator class
  ----------------------------------------------------

    +---------------------------------+
    |  Parent Class: /pif_Iterator    |
    |  Namespace: /pif_Iterator       |
    |  File: MethodIterator class.dm  |
    +---------------------------------+

  The MethodIterator class iterates over the output of a supplied method or function over a domain
  specified by the programmer. This domain can be a list, a numeric range (with an optional step
  size), or another /pif_Iterator object.

    New()
    --------------------

    Arguments.

      The arguments of the MethodIterator constructor consist of the Call portion and the Domain
      portion. That is, it can be written as

        new /pif_Iterator/MethodIterator( [ Call ], [ Domain ] )

      Where the Call portion is one of the following formats:

        1. ProcRef - A path referring to a particular proc. E.g., /prof/Foo or /Foo/proc/Bar.

        2. Object, ProcName
           - Object is an instance of any DM object.
           - ProcName is a string referring to method the Object has.

        3. LibName, FuncName
           - LibName is a the filename of a DLL.
           - FuncName is a string referring to a function in that DLL.

      And the Domain portion is one of the following formats.

        1. List - A /list object

        2. Start, End
           - Start is a number to start the range at.
           - End is a number to end the range at.
           The step size will be one. E.g., the domain will be {Start, Start + 1. Start + 1, ...,
           Start + k, End} though End will only be reached if it is an integer distance from start.

        3. Start, End, StepSize.
           - Start is a number to start the range at.
           - End is a number to end the range at.
           - StepSize is the distance between elements in the range.
           The range will then be {Start, Start + StepSize, Start + 2*StepSize, ..., Start +
           K*StepSize, End}, though End will only be reached if there is a natural number N such
           that End = Start+N*SizeSize.

        4. Iterator - a /pif_Iterator object whose output (specifically, from the Look() function)
           will be used as a domain.

      Behavior.

        Instantiates the MethodIterator class.

      Exceptions.

        - An exception will be thrown if the there are no domain arguments.
        - An exception will be thrown if the constructor can't determine which format the arguments
          are in.

    Done()
    --------------------

    Returns

      True when the end of the data has been reached (i.e., the end of the domain was reached).
      False, otherwise.

    Initialize()
    --------------------

    Arguments.

      Arguments are dependent on the domain iterator.

    Behavior.

      Initializes the MethodIterator object.

    Returns

      True if initialized properly. False, otherwise.

    Next()
    --------------------

    Arguments.

      Arguments are dependent on the domain iterator. If accepted, it will move position in the
      domain to the position specified by the argument.

    Behavior.

      Moves to the next position in the data.

    Returns.

      The position in the domain, or 0 if the end of the domain has been reached.

    Peek()
    --------------------

    Arguments.

      The amount to look forward or backwards in the data.

    Returns..

      The output of the function with the data indicated at the position in the domain.

    Note: Dependent on the implementation in the domain iterator.

    Position()
    --------------------

    Returns.

      The position in the domain.

    Step()
    --------------------

    Arguments.

      The amount to step forwards or backwards in the data.

    Returns.

      The position in the data, or 0 if the new position is out of range.

    Note: Dependent on the implementation of the domain iterator.

  10. Release Notes
  ----------------------------------------------------

  Version 1.02.20191216

    - Reorganized files.
    - Added MIT license to project, under "+License.dm".
    - Updated documentation to reflect MIT license.

  Version 1.01.20151028

    - Added /pif_Iterator/RangeIterator class.
      - Added RangeIterator class documentation.

    - Added /pif_Iterator/MethodIterator class.
      - Added MethodIterator class documentation.

    - Clarified protocol note for when methods are called before Initialize().

  Version 1.0.20151023.

    - Initial release.

 ***************************************************************************************************
 ***************************************************************************************************/